"""
Reads data from Google BigQuery
"""

from logging import getLogger

import google.cloud.bigquery as bigquery
from pandas.core.frame import DataFrame

logger = getLogger(__name__)


class BigQueryClient:
    """Client to read data from Google BigQuery."""

    _bqclient = None

    @property
    def bqclient(self) -> bigquery.Client:
        """Sets up Google BigQuery API Client over the class if not present. If
        it is, returns it.

        :return: The BigQuery's client.
        :rtype: bigquery.Client
        """

        if self._bqclient is None:
            self._bqclient = bigquery.Client()

        return self._bqclient

    def query(
        self, query_string: str, dest_table_id: str, allow_large_results: bool = False
    ) -> DataFrame:
        """Executes query loading into a destination table.

        :param query_string: a query statement
        :type query_string: str
        :param dest_table_id: a fully qualified table as project_id.dataset.table
        :type dest_table_id: str
        :param allow_large_results: allow large query results tables. Defaults to False.
        :type allow_large_results: bool, optional
        :return: The query result.
        :rtype: DataFrame
        """

        job_config = bigquery.QueryJobConfig(destination=dest_table_id)

        result = self.bqclient.query(
            query_string, job_config=job_config, allow_large_results=allow_large_results
        ).result()

        logger.info(f"Query results loaded to the table {dest_table_id}")
        return result



"""
Client used to schedule queries on BigQuery.
"""

import logging
from enum import Enum
from typing import Optional

from google.cloud import bigquery_datatransfer
from google.cloud.bigquery.enums import WriteDisposition

logger = logging.getLogger(__name__)


class DestTableNameTemplate(Enum):
    """Enum used to define BigQuery's table names templates, when applicable."""

    run_date = "_{run_date}"
    run_time = "_{run_time}"
    no_template = ""


class BigQueryDataTransfer:
    """Client used to manipulate BigQuery Data Transfer Service."""

    def __init__(self, project_id: str, service_account: str) -> None:
        """Initializes the class.

        :param project_id: a string containing the Google's project identification
        :type project_id: str
        :param service_account: a string containing a service account e-mail
        :type service_account: str
        """

        self._project_id = project_id
        self._service_account = service_account
        self._bq_datatransfer_client: Optional[
            bigquery_datatransfer.DataTransferServiceClient
        ] = None

    @property
    def bq_datatransfer_client(self) -> bigquery_datatransfer.DataTransferServiceClient:
        """Sets up the BigQuery's DataTransferServiceClient over the class if not present. If
        it is, returns it.

        :return: The Data Transfer Client.
        :rtype: bigquery_datatransfer.DataTransferServiceClient
        """

        if not self._bq_datatransfer_client:
            self._bq_datatransfer_client = (
                bigquery_datatransfer.DataTransferServiceClient()
            )
            self._parent = self._bq_datatransfer_client.common_project_path(
                self._project_id
            )

        return self._bq_datatransfer_client

    def create_or_update_transfer_config(
        self,
        query_str: str,
        dest_dataset_id: str,
        dest_table_id: str,
        squeduled_query_name: str,
        schedule: str,
        dest_table_name_template: str = DestTableNameTemplate.no_template.value,  # noqa: E501
        write_disposition: str = WriteDisposition.WRITE_TRUNCATE,
        partitioning_field: str = "",
    ) -> None:
        """Creates the transfer configuration on Google.

        :param query_str: the SQL statement.
        :type query_str: str
        :param dest_dataset_id: the dataset which will hold the destination data.
        :type dest_dataset_id: str
        :param dest_table_id: the table which will hold the destination data under the
            defined dest_dataset_id.
        :type dest_table_id: str
        :param squeduled_query_name: the query's display name.
        :type squeduled_query_name: str
        :param schedule: string in BigQuery's expected format, such as
            'every mon,wed 09:00'
        :type schedule: str
        :param dest_table_name_template: an optional template to add to a table name,
            useful when using sharded tables. Pass an empty string to ignore,
            defaults to DestTableNameTemplate.no_template.value.
        :type dest_table_name_template: str, optional
        :param partitioning_field: a column name to partition the data. Use empty
            string to not partition the data, defaults to "".
        :type partitioning_field: str, optional
        """

        # List transfers. Store name and display name
        configs = self.bq_datatransfer_client.list_transfer_configs(parent=self._parent)
        config_names = {config.display_name: config.name for config in configs}  # type: ignore # noqa: E501

        # If display name matches one of the already existent display_names, deletes it
        # first by recovering the actual transfer name (which is different from the
        # display name)
        if squeduled_query_name in config_names.keys():
            logger.info(
                f"Deleting old version of scheduled query {squeduled_query_name}."
            )
            already_squeduled_query_name = config_names.get(squeduled_query_name)
            self.bq_datatransfer_client.delete_transfer_config(
                name=already_squeduled_query_name
            )

        # The recreates it (or just create the first time)
        transfer_config = bigquery_datatransfer.TransferConfig(
            destination_dataset_id=dest_dataset_id,
            display_name=squeduled_query_name,
            data_source_id="scheduled_query",
            params={
                "query": query_str,
                "destination_table_name_template": f"{dest_table_id}{dest_table_name_template}",  # noqa: E501
                "write_disposition": write_disposition,
                "partitioning_field": partitioning_field,
            },
            schedule=schedule,
        )

        transfer_config_request = self.bq_datatransfer_client.create_transfer_config(
            bigquery_datatransfer.CreateTransferConfigRequest(
                parent=self._parent,
                transfer_config=transfer_config,
                service_account_name=self._service_account,
            )
        )
        logger.info(f"Created scheduled query {transfer_config_request.name}")



from contextlib import contextmanager
from typing import Iterator, List

from yawl.workflows.base import WorkFlowStep


class Queue:
    """This class alongside the queue context is responsible for defining a
    precedence order of steps first, and finally responsible for calling
    the execute command of each queued step.

    | Usage:

    |     with queue() as q:
    |         q.add(step_1).add(step_2).process()
    """

    def __init__(self) -> None:
        self.__commands: List[WorkFlowStep] = []

    def add(self, command: WorkFlowStep) -> "Queue":
        if self.__commands:
            command.upstream = self.__commands[-1].dest_table  # type: ignore

        self.__commands.append(command)
        return self

    def process(self) -> None:
        for command in self.__commands:
            command.execute()


@contextmanager
def queue() -> Iterator[Queue]:
    """Context that works with a Queue. Refer to Queue class."""

    queue = Queue()
    yield queue



"""
Module used for BigQuery interactions.
"""
import logging
import re
from typing import Optional

from yawl.clients.bigquery_datatransfer import BigQueryDataTransfer, DestTableNameTemplate
from yawl.shared.constants import SERVICE_ACCOUNT_EMAIL
from yawl.workflows.base import WorkFlowStep

logger = logging.getLogger(__name__)


class BigQueryWorkflowStep(WorkFlowStep):
    """This class inherits from the abstract class WorkFlowStep and expands its behavior
    to operate with Google's BigQuery.
    """

    def __init__(
        self, sql: str, dest_table: str, squeduled_query_name: str, schedule: str
    ) -> None:
        """Initializes the class.

        :param sql: a string containing a SQL statement or the path to a .sql file
        :type sql: str
        :param dest_table: contains fully qualified table as "project.dataset.table"
        :type dest_table: str
        :param squeduled_query_name: the scheduled query display name
        :type squeduled_query_name: str
        :param schedule: a string in BigQuery's expected format, such as
            'every mon,wed 09:00'
        :type schedule: str
        """

        self.__sql = sql
        self.__project_id = dest_table.split(".")[0]
        self.__dataset = dest_table.split(".")[1]
        self.__dest_table = dest_table.split(".")[2]
        self.__squeduled_query_name = squeduled_query_name
        self.__schedule = schedule
        self.__upstream = None

        self.__client = BigQueryDataTransfer(
            project_id=self.__project_id,
            service_account=SERVICE_ACCOUNT_EMAIL,  # type: ignore
        )

    @property
    def dest_table(self) -> str:
        """Returns the destination table id.

        :return: The destination table id as string.
        :rtype: str
        """
        return self.__dest_table

    @property
    def upstream(self) -> Optional[str]:
        """Returns the upstream table id.

        :return: The upstream table id as string.
        :rtype: Optional[str]
        """

        return self.__upstream

    @upstream.setter
    def upstream(self, upstream: Optional[str]) -> None:
        self.__upstream = upstream  # type: ignore

    def get_sql_file(self) -> None:
        """Reads a sql file cleaning up extra spaces and new lines making it
        up a single, long, string.
        """

        with open(self.__sql, "r") as f:
            file = f.read()
        self.__sql = re.sub(r"[\s]{2,}|\n", " ", file)

    def execute(self) -> None:
        """Adds the scheduled query into BigQuery."""

        if self.__sql.endswith(".sql"):
            self.get_sql_file()

        self.__client.create_or_update_transfer_config(
            query_str=self.__sql,
            dest_dataset_id=self.__dataset,
            dest_table_id=self.dest_table,
            squeduled_query_name=self.__squeduled_query_name,
            schedule=self.__schedule,
            dest_table_name_template=DestTableNameTemplate.no_template.value,
            write_disposition="WRITE_TRUNCATE",
            partitioning_field="",
        )
        logger.info(
            (
                f"Executing step of upstream {self.upstream}"
                f" and destination {self.dest_table}"
            )
        )



from abc import ABC, abstractmethod
from typing import Optional


class WorkFlowStep(ABC):
    """Abstract class used to define the base of a WorkFlowStep."""

    @abstractmethod
    def execute(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def dest_table(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def upstream(self) -> Optional[str]:
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yawl', 'yawl.clients', 'yawl.shared', 'yawl.workflows']

package_data = \
{'': ['*']}

install_requires = \
['google-cloud-bigquery-datatransfer>=3.3.1,<4.0.0',
 'google-cloud-bigquery-storage==2.0.0',
 'google-cloud-bigquery>=2.6.0,<3.0.0',
 'pandas>=1.3.2,<2.0.0',
 'pytest-xdist>=2.3.0,<3.0.0']

setup_kwargs = {
    'name': 'yawl',
    'version': '0.1.0',
    'description': 'Yet Another WorkLoad - manage scheduled queries [currently] on BigQuery',
    'long_description': '# YAWL - Yet Another Workload\n[\n![Checks](https://img.shields.io/github/checks-status/gbieul/yawl/master)\n![Checks](https://github.com/gbieul/yawl/actions/workflows/yawl-checks.yml/badge.svg)\n![Build](https://github.com/gbieul/yawl/actions/workflows/yawl-build.yml/badge.svg)\n![License](https://img.shields.io/github/license/gbieul/flake8-markdown.svg)\n](https://pypi.org/project/flake8-markdown/)\n\n## 1.0. Intro\n\n## 2.0. Installing YAWL\n\n## 3.0. Using YAWL\n    ',
    'author': 'Gabriel Benvegmi',
    'author_email': 'gbieul_benveg@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/johnfraney/flake8-markdown',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0',
}


setup(**setup_kwargs)



from yawl.workflows.bigquery_workflow import BigQueryWorkflowStep
from yawl.workflows.queue import queue

if __name__ == "__main__":
    step_1 = BigQueryWorkflowStep(
        sql="./sql_files/example.sql",
        dest_table="beaming-ring-330423.transfer_test.table_1",
        squeduled_query_name="test_query101",
        schedule="every mon,wed 09:00",
    )
    step_2 = BigQueryWorkflowStep(
        sql="./sql_files/example.sql",
        dest_table="beaming-ring-330423.transfer_test.table_2",
        squeduled_query_name="test_query_102",
        schedule="every tue,thu 10:00",
    )
    with queue() as q:
        q.add(step_1).add(step_2).process()
